"""
This a companion code to assignment P3 for CSEP576au21
Instructor: Vitaly Ablavsky

The code is based in part on the PyTorch CIFAR-10 tutorial 
and on code in the Scikit-Learn library

Portions of the code that need to be modified or studies as part of this assignment
are marked as :TODO:


Quick-start guide:

(a) python p3.py --mode=unit_test

This will exerciseprecision-recall computation (for you to implement)
The unit test will return False because the function returns placeholder values
    

(b)    python p3.py --mode=train_and_test 
        --cifar10_dir=/path/to/download/cifar-10 
        --outdir=/path/to/visualization_and_models

The first time you run this code, PyTorch will download and unpack CIFAR-10 (about 300Mb)
to the directory you specified and will run training for 1 epoch

(c) Runtime arguments that you can specify on the command line or via your preferred development
    environment
    --n_epochs_train=20 is a good number to try first
    --net_arch=0 is the basline network Net_0 provided for you
    once you implement Net_1 you can specify --net_arch=1



"""

from pdb import set_trace as keyboard
import argparse, sys, os

import matplotlib.pyplot as plt
import numpy as np
from itertools import cycle

import sklearn.metrics 
import sklearn.metrics._ranking

from sklearn.preprocessing import label_binarize


import torch
import torchvision
import torchvision.transforms as transforms

import torch.nn as nn
import torch.nn.functional as F

import torch.optim as optim

rev = 1
student_version = True

fig_fontsize = 14





#########################################################################
#                      Net_0
#########################################################################
class Net_0(nn.Module):
    def __init__(self):
        super().__init__()
        """
        We define Conv2d with these parameters 
          filter size: 5 which implies 5x5 (rows, columns) (our design choice)
          number of input channels/filters: 3 (CIFAR-10 is RGB)
          number of output channels/filters: 6 (our design choice)
          
        For a complete set of parameters of Conv2d see
        https://pytorch.org/docs/stable/generated/torch.nn.Conv2d.html
        ...
        """
        self.conv1 = nn.Conv2d(3, 6, 5) 
        """
        Pool: 
        """
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = nn.Conv2d(6, 16, 5)
        self.fc1 = nn.Linear(16 * 5 * 5, 120)
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, 10)

        self.use_dropout = False


    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = torch.flatten(x, 1) # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


#########################################################################
#                      Net_1
#########################################################################
if not student_version:
    class Net_1(nn.Module):
        pass
else:
    class Net_1(nn.Module):
        """
        :TODO: 
        
        Implement a network with a different architecture:
        try adding or removing a convolutonal layer and changing the number of channels
        You may also change the fully-connected layers.


        Note that when we discussed convolution in class, we talked about several
        strategies for "padding" in order to preserve the input image's shape.
        In most applications of neural nets, we do not include padding (e.g., to avoid
        border effects) and instead let the output be smaller than the input.
        In your experiments you should use this strategy, i.e., use nn.Conv2d with its
        default padding=0.

        You will need to adjust "in_features" for self.fc1 = nn.Linear(in_features,...)
        Although PyTorch makes it easy to find that number via trial-and-error, or
        even avoid the issue altogether by using nn.LazyLinear
        for the purpose of P3 you need to compute in_features by hand and show your work
        The formula is provided here:
        https://pytorch.org/docs/stable/generated/torch.nn.Conv2d.html
        look for "Shape:" and the two equations H_out and W_out

        Is the formula correct?
        """
        pass


#########################################################################
#                      make_network
#########################################################################
def make_network(par):
    """
    Create an instance of a network of known architecture
    """

    net = None
    if par['net_arch'] == 0:
        net = Net_0()
    elif par['net_arch'] == 1:
        net = Net_1()
    else:
        raise ValueError('Unknown network architecture: {}'.format(par['net_arch']))

    return net

#########################################################################
#                      csep576_precision_recall_curve
#########################################################################
def csep576_precision_recall_curve(y_true, y_est):
    """
    _binary_clf_curve() returns three arrays:
    (a) number of false positives (given the score threshold in "thresholds")
    (b) number of true  positives (given the score threshold in "thresholds")
    (c) thresholds sweeping through all the scores

    You should study these variables (e.g., in the debugger or by printing them)
    while calling unit_test to understand what they compute before proceeding
    to precision/recall computations
    """
    fps, tps, thresholds = sklearn.metrics._ranking._binary_clf_curve(
        y_true, y_est) 

    if not student_version:
        pass
    else:
        """
        :TODO: P3 complete this code block. You may want to read
        [CVA2] Ch. 6.3.3 first.
        precision = ...
        recall = ...
        thresholds computed via _binary_clf_curve() should be used as-is

        For now:
        """
        t_ = np.zeros_like(fps)
        return 3*[t_]


    # PR-specific bookkeeping

    precision[np.isnan(precision)] = 0

    # stop when full recall attained
    # and reverse the outputs so recall is decreasing
    last_ind = tps.searchsorted(tps[-1])
    sl = slice(last_ind, None, -1)
    return np.r_[precision[sl], 1], np.r_[recall[sl], 0], thresholds[sl]



#########################################################################
#                      ut_csep576_precision_recall_curve
#########################################################################
def ut_csep576_precision_recall_curve():
    """
    Verify PR computations on several hand-crafted exmples
    """

    print('-'*5 + 'ut_csep576_precision_recall_curve()'+ '-'*5)

    x = np.linspace(-2,2,10)
    y_true = np.ones_like(x).astype('int32')
    y_true[np.where(x < 0)] = 0

    # generate "notional" classifier scores by evaluating a logistic function
    # https://en.wikipedia.org/wiki/Logistic_function
    # y_est = 1/(1 + np.exp(-x)) or a linear function f(x) = a*x
    y_est = x/2 # predictions are in [-1, 1] but that's not required
    
    y_est_0 = np.copy(y_est)

    pr_res_0 = csep576_precision_recall_curve(y_true, y_est_0)

    status = (
        np.array_equal(pr_res_0[0], np.ones_like(pr_res_0[0])) and
        (pr_res_0[1][0] == 1)
        )

    # now introduce a classification error near the boundary
    idx = x.shape[0] //2
    y_est_1 = np.copy(y_est)
    y_est_1[idx] *= -1

    pr_res_1 = csep576_precision_recall_curve(y_true, y_est_1)

    status = (
        status and
        (pr_res_1[0][0] < 1) and
        (pr_res_1[1][0] == 1)
        )

    return status


#########################################################################
#                       viz_precision_recall_multiclass()
#########################################################################
def viz_precision_recall_multiclass(precision, recall, average_precision, par):
    """
    https://scikit-learn.org/stable/auto_examples/model_selection/plot_precision_recall.html#plot-precision-recall-curve-for-each-class-and-iso-f1-curves
    """
    n_classes = len(recall.keys())
    if 'micro' in recall:
        n_classes -= 1

    # setup plot details
    colors_base_seq = ["navy", "turquoise", "darkorange", "cornflowerblue", "teal"]
    colors_cycle = cycle(colors_base_seq)
    n_colors = len(colors_base_seq)
    line_styles = cycle(n_colors*['solid'] + n_colors*['dashed'])
    _, ax = plt.subplots(figsize=(7, 8))

    f_scores = np.linspace(0.2, 0.8, num=4)
    (lines_iso, labels_iso) = plot_iso_F_curves(f_scores)

    if 'micro' in recall and 'micro' in precision:
        display = sklearn.metrics.PrecisionRecallDisplay(
            recall=recall["micro"],
            precision=precision["micro"],
            average_precision=average_precision["micro"],
        )
        display.plot(ax=ax, name="Micro-average precision-recall", color="gold")

    for i, color, line_style in zip(range(n_classes), colors_cycle, line_styles):
        display = sklearn.metrics.PrecisionRecallDisplay(
            recall=recall[i],
            precision=precision[i],
            average_precision=average_precision[i],
        )
        display.plot(ax=ax, name=f"Precision-recall for class {i}", color=color, linestyle=line_style)

    # add the legend for the iso-f1 curves
    handles, labels = display.ax_.get_legend_handles_labels()
    handles.extend(lines_iso)
    labels.extend(labels_iso)
    # set the legend and the axes
    ax.set_xlim([0.0, 1.0])
    ax.set_ylim([0.0, 1.1])
    ax.legend(handles=handles, labels=labels, loc="best")
    if 'pr_fig_title' in par:
        ax.set_title(par['pr_fig_title'])

    plt.show()

    fig_handle = plt.gcf()
    return fig_handle

#########################################################################
#                       plot_iso_F_curves()
#########################################################################
def plot_iso_F_curves(f_scores):
    lines, labels = [], []
    for f_score in f_scores:
        x = np.linspace(0.01, 1)
        y = f_score * x / (2 * x - f_score)
        l, = plt.plot(x[y >= 0], y[y >= 0], color='gray', alpha=0.2)
        plt.annotate('F1={0:0.1f}'.format(f_score), xy=(0.85, y[45] + 0.02), fontsize=fig_fontsize)  

    lines.append(l)
    labels.append('iso-F1 curves')
    return (lines, labels)



#########################################################################
#                       imshow()
#########################################################################
def imshow(img,title=None,suptitle=None):
    img = img / 2 + 0.5     # unnormalize
    npimg = img.numpy()
    plt.clf()
    plt.imshow(np.transpose(npimg, (1, 2, 0)))
    plt.suptitle(suptitle)
    plt.title(title)
    plt.show()





######################################################################################
#                  perform_testing()
######################################################################################
def perform_testing(net, loss_func, dloader, par):
    """
    """
    epoch_loss = 0
    y_est_all = []
    y_gt_all = []
    with torch.no_grad():
        for batch_idx, batch_data in enumerate(dloader):
            images, y_gt = batch_data
            y_est = net(images)

            if loss_func is not None:
                loss = loss_func(y_est, y_gt)
                epoch_loss += loss.detach().item()

            # append predicted and true labels for this batch
            # to the combined lists
            for t_ in [(y_gt_all, y_gt), (y_est_all, y_est)]:
                t_[0].append(t_[1])

    epoch_loss /= batch_idx
    y_est_all = torch.cat(y_est_all)
    y_gt_all = torch.cat(y_gt_all)

    y_est_all = y_est_all.cpu().detach().numpy()
    y_gt_all = y_gt_all.cpu().detach().numpy()

    precision = dict()
    recall = dict()
    avg_precision = dict()
    n_class = y_est_all.shape[1]
    for i in range(n_class):
        y_gt_one_hot = label_binarize(y_gt_all, classes = np.arange(n_class))
        #precision[i], recall[i], thresh_ = sklearn.metrics.precision_recall_curve(y_gt_one_hot[:, i], y_est_all[:, i])

        precision[i], recall[i], thresh_ = csep576_precision_recall_curve(y_gt_one_hot[:, i], y_est_all[:, i])

        avg_precision[i] = sklearn.metrics.average_precision_score(y_gt_one_hot[:, i], y_est_all[:, i])

    return epoch_loss, y_est_all, y_gt_all, precision, recall, avg_precision


######################################################################################
#                  train_and_test()
######################################################################################
def train_and_test(par):
    """

    """

    if 'outdir' not in par or par['outdir'] is None:
        raise IOError('--outdir must be specified')

    outdir = par['outdir']

    if 'cifar10_dir' not in par or par['cifar10_dir'] is None:
        raise IOError('--cifar10_dir must be specified')
    cifar10_dir = par['cifar10_dir']
    if not os.path.exists(cifar10_dir):
        raise IOError('specified cifar10_dir not found: {}'.format(cifar10_dir))



    # Load and normalize CIFAR10, define transforms and data loaders

    transform = transforms.Compose(
        [transforms.ToTensor(),
         transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))])

    batch_size = 4 # needed to construct DataLoader


    trainset = torchvision.datasets.CIFAR10(root=cifar10_dir, train=True,
                                            download=True, transform=transform)
    trainloader = torch.utils.data.DataLoader(trainset, batch_size=batch_size,
                                              shuffle=True, num_workers=0)

    testset = torchvision.datasets.CIFAR10(root=cifar10_dir, train=False,
                                           download=True, transform=transform)
    testloader = torch.utils.data.DataLoader(testset, batch_size=batch_size,
                                             shuffle=False, num_workers=0)


    swap_train_test = False
    if swap_train_test:
        if not student_version:
            pass
        else:
            """
            :TODO: swap the two partitions by
            defining a new set of data loaders; make sure that
            the training data is shuffled (for the SGD) while the test data
            is best (e.g., for debugging) iterated in a fixed order
            """
            pass


    classes = ('plane', 'car', 'bird', 'cat',
               'deer', 'dog', 'frog', 'horse', 'ship', 'truck')




    # Let us show some of the training images, for fun.

    # get some random training images
    dataiter = iter(trainloader)
    images, labels = dataiter.next()

    # print labels
    labels_str = ' '.join('%5s' % classes[labels[j]] for j in range(batch_size))
    print('labels: {}'.format(labels_str))

    # show images
    imshow(torchvision.utils.make_grid(images),title=labels_str)

    viz_filename = 'cifar_tutorial_sample_train_img.png'
    viz_filepath = os.path.join(outdir,viz_filename)
    plt.savefig(viz_filepath)
    print('viz saved to {}'.format(viz_filepath))


    net = make_network(par)

    """
    Define a Loss function and optimizer
    We recommmend a cross-entropy loss 
    https://pytorch.org/docs/stable/generated/torch.nn.CrossEntropyLoss.html

    Does it match the definition in class and in the book?

    and an Stochastic Gradient Descent optimizer
    with a learning rate of 0.001 and momentum 0.9
    https://pytorch.org/docs/stable/generated/torch.optim.SGD.html

    Is the learning rate too conservative? Try a larger value! Does the momentum
    help? Try setting it to zero and see if you notice any difference.
    """

    if not student_version:
        pass
    else:
        class DummyLossValue(object):
            def __init__(self):
                super().__init__()
            def item(self):
                return 0
                
        f_loss = None
        optimizer = None

    """
    You may want to set this value to True to do a quick test
    with the network whose weights are randomly initialized.

    What would you expect to be the accuracy of such a network?
    Does the PR curve match your intuition?
    """
    eval_before_train = False
    if eval_before_train:
        eval_ = perform_testing(net, f_loss, testloader, par)
        fig_title = 'PR_before_training'
        par.update({'pr_fig_title' : fig_title})
        fig_handle = viz_precision_recall_multiclass(eval_[3], eval_[4], eval_[5], par)

        fig_filename = os.path.join(par['outdir'],'{}.pdf'.format(fig_title))
        fig_handle.savefig(fig_filename)
        print('viz saved to {}'.format(fig_filename))
        keyboard()
        plt.close(fig_handle)

    # Train the network
    n_epochs_train = par['n_epochs_train']
    test_loss_history = []
    for epoch in range(n_epochs_train):  # loop over the dataset multiple times
        running_loss = 0.0
        for batch_idx, data in enumerate(trainloader, 0):
            # get the inputs; data is a list of [inputs, labels]
            inputs, labels = data


            if not student_version:
                pass
            else:
                """
                :TODO:
                study the examples here
                https://pytorch.org/docs/stable/optim.html
                then complete the implementation
                """

                # zero the parameter gradients in the optimizer

                # make the forward pass with the network
                # y_est = ....(inputs)
                
                # compute the loss
                
                # backpropagate the gradients
                # hint: use the .backward() method on...what?

                # take a step in the parameter space
                # optimizer.????

                # for now
                loss = DummyLossValue()



            # print statistics
            running_loss += loss.item()
            nb = 1000
            if batch_idx % nb == nb-1:    
                print('[epoch: {:02d}, batch_idx: {:05d}] avg train loss: {:0.3f}'.format(epoch + 1, batch_idx + 1, running_loss / nb))
                running_loss = 0.0
        
        eval_ = perform_testing(net, f_loss, testloader, par)
        #print('Test set loss: {} avg precision: {}'.format(eval_[0], eval_[-1]))
        print('Test set loss: {}'.format(eval_[0]))
        test_loss_history.append(eval_[0])

        fig_title = 'PR_epoch_{:03d}'.format(epoch)
        par.update({'pr_fig_title' : fig_title})
        fig_handle = viz_precision_recall_multiclass(eval_[3], eval_[4], eval_[5], par)

        fig_filename = os.path.join(par['outdir'],'{}.pdf'.format(fig_title))
        fig_handle.savefig(fig_filename)
        plt.close(fig_handle)
        print('viz saved to {}'.format(fig_filename))

        print('end of epoch')
                
    print('Finished Training')
    # An example of how to save a model
    model_PATH = os.path.join(outdir,'cifar_net.pth')
    torch.save(net.state_dict(), model_PATH)
    print('trained model saved to {}'.format(model_PATH))

    # an example of how to load the weights of a model
    net = make_network(par)
    net.load_state_dict(torch.load(model_PATH))
    print('trained model loaded from {}'.format(model_PATH))


    # Test the network on the test data
    dataiter = iter(testloader)
    images, labels = dataiter.next()

    # print images
    gt_str='GroundTruth: ' + ' '.join('%5s' % classes[labels[j]] for j in range(4))
    print(gt_str)


    outputs = net(images)

    _, predicted = torch.max(outputs, 1)

    pred_str = 'Predicted: ' + ' '.join('%5s' % classes[predicted[j]]
                                        for j in range(4))
    print(pred_str)

    imshow(torchvision.utils.make_grid(images),title=gt_str,suptitle=pred_str)
    viz_filename = 'cifar_tutorial_test_results.png'
    viz_filepath = os.path.join(outdir,viz_filename)
    plt.savefig(viz_filepath)
    print('viz saved to {}'.format(viz_filepath))

    print('train_and_test() finished')
    return np.array(test_loss_history)

    
######################################################################################
#                  unit_test()
######################################################################################
def unit_test(par):
    print('-'*5 + 'unit_test'+ '-'*5)

    status = ut_csep576_precision_recall_curve()
    print(status)


######################################################################################
#                  main()
######################################################################################
def main(par):
    mode=par['mode']
    if mode == 'unit_test':
        unit_test(par)
    elif mode == 'train_and_test':
        perf_eval = train_and_test(par)
    else:
        print('no-op: unknown mode {}'.format(mode))

######################################################################################
#                  __main__
######################################################################################
if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--mode', required=True, choices={'unit_test','train_and_test'})
    parser.add_argument('--cifar10_dir', default=None)
    parser.add_argument('--outdir', default=None)
    parser.add_argument('--net_arch', type=int, default=0)
    parser.add_argument('--n_epochs_train', type=int, default=1)
    parser.add_argument('--verbose', action='store_true', default=False)
    
    par__main=vars(parser.parse_args(sys.argv[1:]))
    main(par__main)
